/*
 * Copyright (c) University of Zagreb, Faculty of Electrical Engineering and Computing
 * Authors: Nikola Banic <nikola.banic@fer.hr> and Sven Loncaric <sven.loncaric@fer.hr>
 * 
 * This is only a research code and is therefore only of prototype quality.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * LITERATURE:
 * 
 * H. Yeganeh and Z. Wang, "Objective Quality Assessment of Tone Mapped
 * Images," IEEE Transactios on Image Processing, vol. 22, no. 2, pp. 657- 
 * 667, Feb. 2013.
 * 
 * Tone Mapped image Quality Index (TMQI), Version 1.0
 * Copyright(c) 2012 Hojatollah Yeganeh and Zhou Wang
 * 
 */

#include <cstdio>
#include <cmath>

#if defined(_WIN32) || defined(_WIN64)
#include <cv.h>
#include <highgui.h>
#else
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#endif

#define CV_LOAD_IMAGE_ANYDEPTH 2
#define CV_LOAD_IMAGE_ANYCOLOR 4
#define CV_BGR2GRAY 6
#define CV_BGR2RGB 4

void Gaussian(int size, double sigma, cv::Mat &kernel){
	
	kernel=cv::Mat(size, size, CV_64FC1);
	
	double s=2*sigma*sigma;
	
	double sum=0;
	for (int i=0;i<size;++i){
		for (int j=0;j<size;++j){
			int dx=i-size/2;
			dx*=dx;
			int dy=j-size/2;
			dy*=dy;
			
			double value=exp(-dx/s-dy/s);
			kernel.at<double>(i, j)=value;
			sum+=value;
		}
	}
	kernel/=sum;
}

void RGBtoYxy(cv::Mat source, cv::Mat &destination){
	
	cv::Mat scaled;
	cv::cvtColor(source, scaled, CV_BGR2RGB);
	scaled.convertTo(scaled, CV_64F);
	
	cv::Mat sRGBtoXYZ=(cv::Mat_<double>(3, 3) << 0.4124564,  0.3575761,  0.1804375, 0.2126729,  0.7151522,  0.0721750, 0.0193339, 0.1191920, 0.9503041);
	int rows=source.rows;
	int cols=source.cols;

	cv::Mat reshaped=scaled.reshape(1, rows*cols);
	cv::transpose(reshaped, reshaped);
	cv::Mat xyz=sRGBtoXYZ*reshaped;
	
	cv::transpose(xyz, xyz);
	xyz=xyz.reshape(3, rows);
	
	double eps=2.2204e-16;
	std::vector<cv::Mat> channels;
	cv::split(xyz, channels);
	cv::Mat sum;
	channels[0].copyTo(sum);
	for (int i=1;i<3;++i){
		sum+=channels[i];
	}
	sum+=eps;
	
	cv::Mat x;
	cv::divide(channels[0], sum, x);
	cv::Mat y;
	cv::divide(channels[1], sum, y);
	
	std::vector<cv::Mat> resultChannels;
	resultChannels.push_back(channels[1]);
	resultChannels.push_back(x);
	resultChannels.push_back(y);
	
	cv::Mat result;
	cv::merge(resultChannels, result);
	
	result.copyTo(destination);
}

double normcdf(double x, double mu, double sigma){
    
    x=(x-mu)/sigma;
	
	double a1=0.254829592;
    double a2=-0.284496736;
    double a3=1.421413741;
    double a4=-1.453152027;
    double a5=1.061405429;
    double p=0.3275911;

    int sign=1;
    if (x<0){
        sign=-1;
    }
    x=fabs(x)/sqrt(2.0);

    double t=1.0/(1.0+p*x);
    double y=1.0-(((((a5*t+a4)*t)+a3)*t+a2)*t+a1)*t*exp(-x*x);

    return 0.5*(1.0+sign*y);
}

double Slocal(cv::Mat L_hdr, cv::Mat L_ldr, cv::Mat window, double sf){
	
	double C1=0.01;
	double C2=10;
	
	double windowSum=0;
	for (int i=0;i<window.rows*window.cols;++i){
		windowSum+=*(((double *)window.data)+i);
	}
	window/=windowSum;
	
	cv::Mat img1;
	L_hdr.convertTo(img1, CV_64F);
	cv::Mat img2;
	L_ldr.convertTo(img2, CV_64F);
	
	cv::Mat mu1;
	cv::filter2D(img1, mu1, -1, window);
	cv::Mat mu2;
	cv::filter2D(img2, mu2, -1, window);
	
	cv::Mat mu1_sq=mu1.mul(mu1);
	cv::Mat mu2_sq=mu2.mul(mu2);
	cv::Mat mu1_mu2=mu1.mul(mu2);
	
	cv::Mat img1_2=img1.mul(img1);
	
	cv::Mat img2_2=img2.mul(img2);
	
	cv::Mat img1_img2=img1.mul(img2);
	
	cv::Mat sigma1_sq;
	cv::filter2D(img1_2, sigma1_sq, -1, window);
	sigma1_sq-=mu1_sq;
	
	cv::Mat sigma2_sq;
	cv::filter2D(img2_2, sigma2_sq, -1, window);
	sigma2_sq-=mu2_sq;
	
	cv::Mat sigma1;
	sigma1_sq.copyTo(sigma1);
	for (int i=0;i<sigma1.rows*sigma1.cols;++i){
		double &value=*(((double *)sigma1.data)+i);
		if (value<0){
			value=0;
		}
		value=sqrt(value);
	}
	
	cv::Mat sigma2;
	sigma2_sq.copyTo(sigma2);
	for (int i=0;i<sigma2.rows*sigma2.cols;++i){
		double &value=*(((double *)sigma2.data)+i);
		if (value<0){
			value=0;
		}
		value=sqrt(value);
	}
	
	cv::Mat sigma12;
	cv::filter2D(img1_img2, sigma12, -1, window);
	sigma12-=mu1_mu2;
	
	double CSF=100.*2.6*(0.0192+0.114*sf)*exp(-pow((0.114*sf), 1.1));
	
	double u_hdr=128/(1.4*CSF);
	double sig_hdr=u_hdr/3;
	cv::Mat sigma1p;
	sigma1.copyTo(sigma1p);
	for (int i=0;i<sigma1p.rows*sigma1p.cols;++i){
		double &value=*(((double *)(sigma1p.data))+i);
		value=normcdf(value, u_hdr, sig_hdr);
	}
	
	double u_ldr=u_hdr;
	double sig_ldr=u_ldr/3;
	cv::Mat sigma2p;
	sigma2.copyTo(sigma2p);
	for (int i=0;i<sigma2p.rows*sigma2p.cols;++i){
		double &value=*(((double *)(sigma2p.data))+i);
		value=normcdf(value, u_ldr, sig_ldr);
	}
	
	cv::Mat s_map=sigma1p.mul(sigma2p);
	s_map=2*s_map+C1;
	
	cv::Mat aa=sigma1p.mul(sigma1p);
	cv::Mat bb=sigma2p.mul(sigma2p);
	
	cv::Mat b=aa+bb+C1;
	
	cv::divide(s_map, b, s_map);
	
	aa=sigma12+C2;
	bb=sigma1.mul(sigma2);
	bb+=C2;
	cv::divide(aa, bb, b);
	
	s_map=s_map.mul(b);
	
	double s=0.0;
	for (int i=0;i<s_map.rows*s_map.cols;++i){
		s+=*(((double *)s_map.data)+i);
	}
	
	s/=s_map.rows*s_map.cols;
	
	return s;
}

double StructuralFidelity(cv::Mat L_hdr_input, cv::Mat L_ldr_input, int level, double weight[], cv::Mat window){
	
	cv::Mat downsample_filter=cv::Mat::ones(2, 2, CV_64FC1)/4.0;
	
	cv::Mat L_hdr;
	cv::Mat L_ldr;
	L_hdr_input.convertTo(L_hdr, CV_64F);
	L_ldr_input.convertTo(L_ldr, CV_64F);
	
	double *s_local=new double[level];
	int f=32;
	for (int l=1;l<=level;++l){
		f/=2;
		s_local[l-1]=Slocal(L_hdr, L_ldr, window, f);
		
		cv::Mat a=cv::Mat::zeros(L_hdr.rows/2, L_hdr.cols/2, CV_64FC1);
		cv::Mat b=cv::Mat::zeros(L_ldr.rows/2, L_ldr.cols/2, CV_64FC1);
		
		for (int i=0;i<a.rows;++i){
			for (int j=0;j<a.cols;++j){
				for (int ii=0;ii<2;++ii){
					for (int jj=0;jj<2;++jj){
						*(((double *)a.data)+i*a.cols+j)+=*(((double *)L_hdr.data)+(2*i+ii)*L_hdr.cols+2*j+jj);
					}
				}
				*(((double *)a.data)+i*a.cols+j)/=4;
			}
		}
		
		for (int i=0;i<b.rows;++i){
			for (int j=0;j<b.cols;++j){
				for (int ii=0;ii<2;++ii){
					for (int jj=0;jj<2;++jj){
						*(((double *)b.data)+i*b.cols+j)+=*(((double *)L_ldr.data)+(2*i+ii)*L_ldr.cols+2*j+jj);
					}
				}
				*(((double *)b.data)+i*b.cols+j)/=4;
			}
		}
		
		a.copyTo(L_hdr);
		b.copyTo(L_ldr);
		
	}
	
	double s=1.0;
	for (int i=0;i<level;++i){
		s*=pow(s_local[i], weight[i]);
	}
	
	delete[] s_local;
	
	return s;
}

double betapdf(double x, double a, double b){
   if (x < 0 || x > 1) return 0;
   double tmp = pow((1 - x), (b - 1)) * pow(x, (a - 1));
   double iB = exp(lgamma(a + b) - lgamma(a) - lgamma(b));
   return tmp * iB;
}

double normpdf(double x, double mu, double sigma){
	
	return (1.0/(sigma*sqrt(2.0*M_PI)))*exp(-(x-mu)*(x-mu)/(2.0*sigma*sigma));
}

double StatisticalNaturalness(cv::Mat L_ldr){
	
	double u=0.0;
	for (int i=0;i<L_ldr.rows*L_ldr.cols;++i){
		u+=*(((double *)L_ldr.data)+i);
	}
	u/=L_ldr.rows*L_ldr.cols;
	
	int rr=L_ldr.rows/11;
	int cc=L_ldr.cols/11;
	if (L_ldr.rows%11!=0){
		++rr;
	}
	if (L_ldr.cols%11!=0){
		++cc;
	}
	
	double sig=0.0;
	int allCount=0;
	for (int i=0;i<rr;++i){
		for (int j=0;j<cc;++j){
			
			double m=0.0;
			
			int count=0;
			for (int ii=11*i;ii<11*(i+1);++ii){
				for (int jj=11*j;jj<11*(j+1);++jj){
					double value=0;
					if (ii<L_ldr.rows && jj<L_ldr.cols){
						value+=*(((double *)L_ldr.data)+ii*L_ldr.cols+jj);
						++count;
					}
					m+=value;
				}
			}
			
			m/=11*11;
			
			double s=0.0;
			
			for (int ii=11*i;ii<11*(i+1);++ii){
				for (int jj=11*j;jj<11*(j+1);++jj){
					double value=0;
					if (ii<L_ldr.rows && jj<L_ldr.cols){
						value+=*(((double *)L_ldr.data)+ii*L_ldr.cols+jj);
					}
					s+=(value-m)*(value-m);
				}
			}
			
			s/=11*11-1;
			s=sqrt(s);
			sig+=count*s;
			allCount+=count;
		}
	}
	sig/=allCount;
	
	double phat_1=4.4;
	double phat_2=10.1;
	double beta_mode=(phat_1-1)/(phat_1+phat_2-2);
	double C_0=betapdf(beta_mode, phat_1, phat_2);
	double C=betapdf(sig/64.29, phat_1, phat_2);
	double pc=C/C_0;
	
	double muhat=115.94;
	double sigmahat=27.99;
	double B=normpdf(u, muhat, sigmahat);
	double B_0=normpdf(muhat, muhat, sigmahat);
	double pb=B/B_0;
	
	double N=pb*pc;
	
	return N;
}

double TMQI(cv::Mat hdrImage, cv::Mat ldrImage, cv::Mat window){
	
	double a=0.8012;
	double alpha=0.3046;
	double beta=0.7088;
	
	int level=5;
	double weight[]={0.0448, 0.2856, 0.3001, 0.2363, 0.1333};
	
	cv::Mat HDR;
	RGBtoYxy(hdrImage, HDR);
	std::vector<cv::Mat> hdrChannels;
	cv::split(HDR, hdrChannels);
	cv::Mat L_hdr;
	hdrChannels[0].copyTo(L_hdr);
	
	double lmin, lmax;
	minMaxLoc(L_hdr, &lmin, &lmax);
	L_hdr=(int)(((1ll<<32)-1)/(lmax - lmin)+0.5)*(L_hdr-lmin);
	L_hdr.convertTo(L_hdr, CV_64F);
	
	
	cv::Mat LDR;
	RGBtoYxy(ldrImage, LDR);
	std::vector<cv::Mat> ldrChannels;
	cv::split(LDR, ldrChannels);
	cv::Mat L_ldr;
	ldrChannels[0].copyTo(L_ldr);
	
	double structuralFidelity=StructuralFidelity(L_hdr, L_ldr, level, weight, window);
	
	double naturalness=StatisticalNaturalness(L_ldr);
	
	double tmqi=a*pow(structuralFidelity, alpha)+(1-a)*pow(naturalness, beta);
	
	return tmqi;
}

double TMQI(cv::Mat hdrImage, cv::Mat ldrImage){
	cv::Mat window;
	Gaussian(11, 1.5, window);
	return TMQI(hdrImage, ldrImage, window);
}

int main(int argc, char **argv){
	
	if (argc<3){
		printf("Usage: %s HDR_file LDR_file [factor]\n", argv[0]);
		printf("\tfactor - the factor that the HDR image is multiplied by (1 by default)\n");
		return 0;
	}
	
	cv::Mat img1=cv::imread(argv[1], CV_LOAD_IMAGE_ANYDEPTH+CV_LOAD_IMAGE_ANYCOLOR);
	cv::Mat img2=cv::imread(argv[2], CV_LOAD_IMAGE_ANYDEPTH+CV_LOAD_IMAGE_ANYCOLOR);
	
	double factor=1.0;
	
	if (argc>3){
		sscanf(argv[3], "%lf", &factor);
	}
	
	double tmqi=TMQI(factor*img1, img2);
	
	printf("%lf\n", tmqi);
	
	return 0;
}
